"""
Implements Supply-Chain Firewall's `configure` subcommand.
"""

from argparse import Namespace
import logging

from scfw.configure.constants import *  # noqa
import scfw.configure.dd_agent as dd_agent
import scfw.configure.env as env
import scfw.configure.interactive as interactive
from scfw.configure.interactive import GREETING

_log = logging.getLogger(__name__)


def run_configure(args: Namespace) -> int:
    """
    Configure the environment for use with the supply-chain firewall.

    Args:
        args: A `Namespace` containing the parsed `configure` subcommand command line.

    Returns:
        An integer status code, 0 or 1.
    """
    try:
        if args.remove:
            # These options result in the firewall's configuration block being removed
            env.update_config_files({
                "alias_pip": False,
                "alias_npm": False,
                "dd_agent_port": None,
                "dd_api_key": None,
                "dd_log_level": None
            })
            dd_agent.remove_agent_logging()
            print(
                "All Supply-Chain Firewall-managed configuration has been removed from your environment."
                "\n\nPost-removal tasks:"
                "\n* Update your current shell environment by sourcing from your .bashrc/.zshrc file."
                "\n* If you had previously configured Datadog Agent log forwarding, restart the Agent."
            )
            return 0

        # The CLI parser guarantees that all of these arguments are present
        is_interactive = not any(
            {args.alias_pip, args.alias_npm, args.dd_agent_port, args.dd_api_key, args.dd_log_level}
        )

        if is_interactive:
            print(GREETING)
            answers = interactive.get_answers()
        else:
            answers = vars(args)

        if not answers:
            return 0

        if (port := answers.get("dd_agent_port")):
            dd_agent.configure_agent_logging(port)

        env.update_config_files(answers)

        if is_interactive:
            print(interactive.get_farewell(answers))

        return 0

    except Exception as e:
        _log.error(e)
        return 1



"""
Provides a `FirewallLogger` class for sending logs to Datadog.
"""

import json
import logging
import os

import dotenv

import scfw
from scfw.configure import DD_ENV, DD_LOG_LEVEL_VAR, DD_SERVICE, DD_SOURCE
from scfw.ecosystem import ECOSYSTEM
from scfw.logger import FirewallAction, FirewallLogger
from scfw.target import InstallTarget

_log = logging.getLogger(__name__)

_DD_LOG_LEVEL_DEFAULT = FirewallAction.BLOCK


dotenv.load_dotenv()


class DDLogFormatter(logging.Formatter):
    """
    A custom JSON formatter for firewall logs.
    """
    def format(self, record) -> str:
        """
        Format a log record as a JSON string.

        Args:
            record: The log record to be formatted.
        """
        log_record = {
            "source": DD_SOURCE,
            "service": DD_SERVICE,
            "version": scfw.__version__,
            "env": os.getenv("DD_ENV", DD_ENV),
        }

        for key in {"action", "created", "ecosystem", "msg", "targets"}:
            log_record[key] = record.__dict__[key]

        return json.dumps(log_record) + '\n'


class DDLogger(FirewallLogger):
    """
    An implementation of `FirewallLogger` for sending logs to Datadog.
    """
    def __init__(self, logger: logging.Logger):
        """
        Initialize a new `DDLogger`.

        Args:
            logger: A configured log handle to which logs will be written.
        """
        self._logger = logger

        try:
            self._level = FirewallAction(os.getenv(DD_LOG_LEVEL_VAR))
        except ValueError:
            _log.warning(f"Undefined or invalid Datadog log level: using default level {_DD_LOG_LEVEL_DEFAULT}")
            self._level = _DD_LOG_LEVEL_DEFAULT

    def log(
        self,
        action: FirewallAction,
        ecosystem: ECOSYSTEM,
        command: list[str],
        targets: list[InstallTarget]
    ):
        """
        Receive and log data about a completed firewall run.

        Args:
            action: The action taken by the firewall.
            ecosystem: The ecosystem of the inspected package manager command.
            command: The package manager command line provided to the firewall.
            targets: The installation targets relevant to firewall's action.
        """
        if not self._level or action < self._level:
            return

        self._logger.info(
            f"Command '{' '.join(command)}' was {str(action).lower()}ed",
            extra={"action": str(action), "ecosystem": str(ecosystem), "targets": list(map(str, targets))}
        )



"""
Provides utilities for interactively accepting configuration options from the user.
"""

import os

import inquirer  # type: ignore

from scfw.configure.constants import DD_API_KEY_VAR
from scfw.logger import FirewallAction

GREETING = (
    "Thank you for using scfw, the Supply-Chain Firewall by Datadog!\n\n"
    "scfw is a tool for preventing the installation of malicious PyPI and npm packages.\n\n"
    "This script will walk you through setting up your environment to get the most out\n"
    "of scfw. You can rerun this script at any time.\n"
)

_DD_AGENT_DEFAULT_LOG_PORT = "10365"


def get_answers() -> dict:
    """
    Get the user's selection of configuration options in interactive mode.

    Returns:
        A `dict` containing the user's selected configuration options.
    """
    has_dd_api_key = os.getenv(DD_API_KEY_VAR) is not None

    questions = [
        inquirer.Confirm(
            name="alias_pip",
            message="Would you like to set a shell alias to run all pip commands through the firewall?",
            default=True
        ),
        inquirer.Confirm(
            name="alias_npm",
            message="Would you like to set a shell alias to run all npm commands through the firewall?",
            default=True
        ),
        inquirer.Confirm(
            name="dd_agent_logging",
            message="If you have the Datadog Agent installed locally, would you like to forward firewall logs to it?",
            default=False
        ),
        inquirer.Text(
            name="dd_agent_port",
            message=f"Enter the local port where the Agent will receive logs (default: {_DD_AGENT_DEFAULT_LOG_PORT})",
            ignore=lambda answers: not answers["dd_agent_logging"]
        ),
        inquirer.Confirm(
            name="dd_api_logging",
            message="Would you like to enable sending firewall logs to Datadog using an API key?",
            default=False,
            ignore=lambda answers: has_dd_api_key or answers["dd_agent_logging"]
        ),
        inquirer.Text(
            name="dd_api_key",
            message="Enter a Datadog API key",
            validate=lambda _, current: current != '',
            ignore=lambda answers: has_dd_api_key or not answers["dd_api_logging"]
        ),
        inquirer.List(
            name="dd_log_level",
            message="Select the desired log level for Datadog logging",
            choices=[str(action) for action in FirewallAction],
            ignore=lambda answers: not (answers["dd_agent_logging"] or has_dd_api_key or answers["dd_api_logging"])
        )
    ]

    answers = inquirer.prompt(questions)

    # Patch for inquirer's broken `default` option
    if answers.get("dd_agent_logging") and not answers.get("dd_agent_port"):
        answers["dd_agent_port"] = _DD_AGENT_DEFAULT_LOG_PORT

    return answers


def get_farewell(answers: dict) -> str:
    """
    Generate a farewell message in interactive mode based on the configuration
    options selected by the user.

    Args:
        answers: The dictionary of user-selected configuration options.

    Returns:
        A `str` farewell message to print in interactive mode.
    """
    farewell = (
        "The environment was successfully configured for Supply-Chain Firewall."
        "\n\nPost-configuration tasks:"
        "\n* Update your current shell environment by sourcing from your .bashrc/.zshrc file."
    )

    if answers.get("dd_agent_logging"):
        farewell += "\n* Restart the Datadog Agent in order for it to accept firewall logs."

    farewell += "\n\nGood luck!"

    return farewell



"""
Implements the supply-chain firewall's core `run` subcommand.
"""

from argparse import Namespace
import inquirer  # type: ignore
import logging

from scfw.command import UnsupportedVersionError
import scfw.commands as commands
from scfw.ecosystem import ECOSYSTEM
from scfw.logger import FirewallAction, FirewallLogger
import scfw.loggers as loggers
from scfw.verifier import FindingSeverity
import scfw.verifiers as verifs
import scfw.verify as verify
from scfw.target import InstallTarget

_log = logging.getLogger(__name__)


def run_firewall(args: Namespace) -> int:
    """
    Run a package manager command through the supply-chain firewall.

    Args:
        args:
            A `Namespace` parsed from a `run` subcommand command line containing
            a `command` to run through the firewall.
        help: A help message to print in the case of early returns.

    Returns:
        An integer status code, 0 or 1.
    """
    try:
        logs = loggers.get_firewall_loggers()
        _log.info(f"Command: '{' '.join(args.command)}'")

        ecosystem, command = commands.get_package_manager_command(args.command, executable=args.executable)
        targets = command.would_install()
        _log.info(f"Command would install: [{', '.join(map(str, targets))}]")

        if targets:
            verifiers = verifs.get_install_target_verifiers()
            _log.info(
                f"Using installation target verifiers: [{', '.join(v.name() for v in verifiers)}]"
            )

            reports = verify.verify_install_targets(verifiers, targets)

            if (critical_report := reports.get(FindingSeverity.CRITICAL)):
                _log_firewall_action(
                    logs,
                    FirewallAction.BLOCK,
                    ecosystem,
                    args.command,
                    list(critical_report)
                )
                print(verify.show_verification_report(critical_report))
                print("\nThe installation request was blocked. No changes have been made.")
                return 0

            if (warning_report := reports.get(FindingSeverity.WARNING)):
                print(verify.show_verification_report(warning_report))
                if not (inquirer.confirm("Proceed with installation?", default=False)):
                    _log_firewall_action(
                        logs,
                        FirewallAction.ABORT,
                        ecosystem,
                        args.command,
                        list(warning_report)
                    )
                    print("The installation request was aborted. No changes have been made.")
                    return 0

        if args.dry_run:
            _log.info("Firewall dry-run mode enabled: command will not be run")
            print("Dry-run: exiting without running command.")
        else:
            _log_firewall_action(
                logs,
                FirewallAction.ALLOW,
                ecosystem,
                args.command,
                targets
            )
            command.run()
        return 0

    except UnsupportedVersionError as e:
        _log.error(f"Incompatible package manager version: {e}")
        return 0

    except Exception as e:
        _log.error(e)
        return 1


def _log_firewall_action(
    logs: list[FirewallLogger],
    action: FirewallAction,
    ecosystem: ECOSYSTEM,
    command: list[str],
    targets: list[InstallTarget],
):
    """
    Log a firewall action across a given set of client loggers.

    Args:
        action: The action taken by the firewall.
        ecosystem: The ecosystem of the inspected package manager command.
        command: The package manager command line provided to the firewall.
        targets:
            The installation targets relevant to firewall's action.
    """
    # One would like to use `map` for this, but it is lazily evaluated
    for log in logs:
        log.log(action, ecosystem, command, targets)



"""
Defines the supply-chain firewall's command-line interface and performs argument parsing.
"""

from argparse import ArgumentError, Namespace
from enum import Enum
import logging
import sys
from typing import Callable, Optional

import scfw
from scfw.ecosystem import ECOSYSTEM
from scfw.logger import FirewallAction
from scfw.parser import ArgumentParser

_LOG_LEVELS = list(
    map(
        logging.getLevelName,
        [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR]
    )
)
_DEFAULT_LOG_LEVEL = logging.getLevelName(logging.WARNING)


def _add_configure_cli(parser: ArgumentParser):
    """
    Defines the command-line interface for the firewall's `configure` subcommand.

    Args:
        parser: The `ArgumentParser` to which the `configure` command line will be added.
    """
    parser.add_argument(
        "-r",
        "--remove",
        action="store_true",
        help="Remove all Supply-Chain Firewall-managed configuration"
    )

    parser.add_argument(
        "--alias-pip",
        action="store_true",
        help="Add shell aliases to always run pip commands through Supply-Chain Firewall"
    )

    parser.add_argument(
        "--alias-npm",
        action="store_true",
        help="Add shell aliases to always run npm commands through Supply-Chain Firewall"
    )

    parser.add_argument(
        "--dd-agent-port",
        type=str,
        default=None,
        metavar="PORT",
        help="Configure log forwarding to the local Datadog Agent on the given port"
    )

    parser.add_argument(
        "--dd-api-key",
        type=str,
        default=None,
        metavar="KEY",
        help="API key to use when forwarding logs via the Datadog API"
    )

    parser.add_argument(
        "--dd-log-level",
        type=str,
        default=None,
        choices=[str(action) for action in FirewallAction],
        metavar="LEVEL",
        help="Desired logging level for Datadog log forwarding (options: %(choices)s)"
    )


def _add_run_cli(parser: ArgumentParser):
    """
    Defines the command-line interface for the firewall's `run` subcommand.

    Args:
        parser: The `ArgumentParser` to which the `run` command line will be added.
    """
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Verify any installation targets but do not run the package manager command"
    )

    parser.add_argument(
        "--executable",
        type=str,
        default=None,
        metavar="PATH",
        help="Python or npm executable to use for running commands (default: environmentally determined)"
    )


class Subcommand(Enum):
    """
    The set of subcommands that comprise the supply-chain firewall's command line.
    """
    Configure = "configure"
    Run = "run"

    def __str__(self) -> str:
        """
        Format a `Subcommand` for printing.

        Returns:
            A `str` representing the given `Subcommand` suitable for printing.
        """
        return self.value

    def _parser_spec(self) -> dict:
        """
        Return the `ArgumentParser` configuration for the given subcommand's parser.

        Returns:
            A `dict` of `kwargs` to pass to the `argparse.SubParsersAction.add_parser()`
            method for configuring the subparser corresponding to the subcommand.
        """
        match self:
            case Subcommand.Configure:
                return {
                    "exit_on_error": False,
                    "description": "Configure the environment for using Supply-Chain Firewall."
                }
            case Subcommand.Run:
                return {
                    "usage": "%(prog)s [options] COMMAND",
                    "exit_on_error": False,
                    "description": "Run a package manager command through Supply-Chain Firewall."
                }

    def _cli_spec(self) -> Callable[[ArgumentParser], None]:
        """
        Return a function for adding the given subcommand's command-line options
        to a given `ArgumentParser`.

        Returns:
            A `Callable[[ArgumentParser], None]` that adds the command-line options
            for the subcommand to the `ArgumentParser` it is given, in the intended
            case via a sequence of calls to `ArgumentParser.add_argument()`.
        """
        match self:
            case Subcommand.Configure:
                return _add_configure_cli
            case Subcommand.Run:
                return _add_run_cli


def _cli() -> ArgumentParser:
    """
    Defines the command-line interface for the supply-chain firewall.

    Returns:
        A parser for the supply-chain firewall's command line.

        This parser only handles the firewall's own optional arguments and subcommands.
        It does not parse the package manager commands being run through the firewall.
    """
    parser = ArgumentParser(
        prog="scfw",
        exit_on_error=False,
        description="A tool for preventing the installation of malicious PyPI and npm packages."
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=scfw.__version__
    )

    parser.add_argument(
        "--log-level",
        type=str,
        choices=_LOG_LEVELS,
        default=_DEFAULT_LOG_LEVEL,
        metavar="LEVEL",
        help="Desired logging level (default: %(default)s, options: %(choices)s)"
    )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    for subcommand in Subcommand:
        subparser = subparsers.add_parser(str(subcommand), **subcommand._parser_spec())
        subcommand._cli_spec()(subparser)

    return parser


def _parse_command_line(argv: list[str]) -> tuple[Optional[Namespace], str]:
    """
    Parse the supply-chain firewall's command line from a given argument vector.

    Args:
        argv: The argument vector to be parsed.

    Returns:
        A `tuple` of a `Namespace` object containing the results of parsing the given
        argument vector and a `str` help message for the caller's use in early exits.
        In the case of a parsing failure, `None` is returned instead of a `Namespace`.

        On success, and only for the `run` subcommand, the returned `Namespace` contains
        the package manager command present in the given argument vector as a `list[str]`
        under the `command` attribute.
    """
    hinge = len(argv)
    for ecosystem in ECOSYSTEM:
        try:
            hinge = min(hinge, argv.index(str(ecosystem)))
        except ValueError:
            pass

    parser = _cli()
    help_msg = parser.format_help()

    try:
        args = parser.parse_args(argv[1:hinge])

        # Config removal option is mutually exclusive with the others
        if (
            Subcommand(args.subcommand) == Subcommand.Configure
            and args.remove
            and any({args.alias_pip, args.alias_npm, args.dd_agent_port, args.dd_api_key, args.dd_log_level})
        ):
            raise ArgumentError(None, "Cannot combine configuration and removal options")

        # Only allow a package manager `command` argument when the user selected
        # the `run` subcommand
        match Subcommand(args.subcommand), argv[hinge:]:
            case Subcommand.Run, []:
                raise ArgumentError(None, "Missing required package manager command")
            case Subcommand.Run, _:
                args.command = argv[hinge:]
            case _, []:
                pass
            case _:
                raise ArgumentError(None, "Received unexpected package manager command")

        return args, help_msg

    except ArgumentError:
        return None, help_msg


def parse_command_line() -> tuple[Optional[Namespace], str]:
    """
    Parse the supply-chain firewall's command line.

    Returns:
        A `tuple` of a `Namespace` object containing the results of parsing the
        firewall's command line and a `str` help message for the caller's use in
        early exits. In the case of a parsing failure, `None` is returned instead
        of a `Namespace`.

        On success, the returned `Namespace` contains the package manager command
        provided to the firewall as a (possibly empty) `list[str]` under the `command`
        attribute.
    """
    return _parse_command_line(sys.argv)



"""
Provides an interface for client loggers to receive information about a
completed run of the supply-chain firewall.
"""

from abc import (ABCMeta, abstractmethod)
from enum import Enum

from scfw.ecosystem import ECOSYSTEM
from scfw.target import InstallTarget


class FirewallAction(Enum):
    """
    The various actions the firewall may take in response to inspecting a
    package manager command.
    """
    ALLOW = "ALLOW"
    ABORT = "ABORT"
    BLOCK = "BLOCK"

    def __lt__(self, other) -> bool:
        """
        Compare two `FirewallAction` instances on the basis of their underlying numeric values.

        Args:
            self: The `FirewallAction` to be compared on the left-hand side
            other: The `FirewallAction` to be compared on the right-hand side

        Returns:
            A `bool` indicating whether `<` holds between the two given `FirewallAction`.

        Raises:
            TypeError: The other argument given was not a `FirewallAction`.
        """
        if self.__class__ is not other.__class__:
            raise TypeError(
                f"'<' not supported between instances of '{self.__class__}' and '{other.__class__}'"
            )

        match self.name, other.name:
            case "ALLOW", "ABORT":
                return True
            case "ALLOW", "BLOCK":
                return True
            case "ABORT", "BLOCK":
                return True
            case _:
                return False

    def __str__(self) -> str:
        """
        Format a `FirewallAction` for printing.

        Returns:
            A `str` representing the given `FirewallAction` suitable for printing.
        """
        return self.value


class FirewallLogger(metaclass=ABCMeta):
    """
    An interface for passing information about a completed firewall run to
    client loggers.
    """
    @abstractmethod
    def log(
        self,
        action: FirewallAction,
        ecosystem: ECOSYSTEM,
        command: list[str],
        targets: list[InstallTarget]
    ):
        """
        Pass data from a completed run of the firewall to a logger.

        Args:
            action: The action taken by the firewall.
            ecosystem: The ecosystem of the inspected package manager command.
            command: The package manager command line provided to the firewall.
            targets:
                The installation targets relevant to firewall's action.

                In the case of a blocking action, `targets` contains the installation
                targets that caused the firewall to block.  In the case of an aborting
                action, `targets` contains the targets that prompted the firewall to
                warn the user and seek confirmation to proceed.
        """
        pass



"""
Defines a subclass of `PackageManagerCommand` for `npm` commands.
"""

import logging
import subprocess
from typing import Optional

from packaging.version import InvalidVersion, Version, parse as version_parse

from scfw.command import PackageManagerCommand, UnsupportedVersionError
from scfw.ecosystem import ECOSYSTEM
from scfw.target import InstallTarget

_log = logging.getLogger(__name__)

MIN_NPM_VERSION = version_parse("7.0.0")

_UNSUPPORTED_NPM_VERSION = f"npm before v{MIN_NPM_VERSION} is not supported"

# The "placeDep" log lines describe a new dependency added to the
# dependency tree being constructed by an installish command
_NPM_LOG_PLACE_DEP = "placeDep"

# Each added dependency is always the fifth token in its log line
_NPM_LOG_DEP_TOKEN = 4


class NpmCommand(PackageManagerCommand):
    """
    A representation of `npm` commands via the `PackageManagerCommand` interface.
    """
    def __init__(self, command: list[str], executable: Optional[str] = None):
        """
        Initialize a new `NpmCommand`.

        Args:
            command: An `npm` command line.
            executable:
                Optional path to the executable to run the command.  Determined by the
                environment if not given.

        Raises:
            ValueError: An invalid `npm` command was given.
            UnsupportedVersionError:
                An unsupported version of `npm` was used to initialize an `NpmCommand`.
        """
        def get_npm_version(executable) -> Version:
            try:
                # All supported versions adhere to this format
                npm_version_command = [executable, "--version"]
                version_str = subprocess.run(npm_version_command, check=True, text=True, capture_output=True)
                return version_parse(version_str.stdout.strip())
            except InvalidVersion:
                raise UnsupportedVersionError(_UNSUPPORTED_NPM_VERSION)

        if not command or command[0] != "npm":
            raise ValueError("Malformed npm command")
        self._command = command
        self._executable = "npm"

        if executable:
            self._command[0] = self._executable = executable
        if get_npm_version(self._executable) < MIN_NPM_VERSION:
            raise UnsupportedVersionError(_UNSUPPORTED_NPM_VERSION)

    def run(self):
        """
        Run an `npm` command.
        """
        subprocess.run(self._command)

    def would_install(self) -> list[InstallTarget]:
        """
        Determine the list of packages an `npm` command would install if it were run.

        Returns:
            A `list[InstallTarget]` representing the packages the `npm` command would
            install if it were run.

        Raises:
            ValueError: The `npm` dry-run output does not have the expected format.
        """
        def is_place_dep_line(line: str) -> bool:
            return _NPM_LOG_PLACE_DEP in line

        def line_to_dependency(line: str) -> str:
            return line.split()[_NPM_LOG_DEP_TOKEN]

        def str_to_install_target(s: str) -> InstallTarget:
            package, sep, version = s.rpartition('@')
            if version == s or (sep and not package):
                raise ValueError("Failed to parse npm install target")
            return InstallTarget(ECOSYSTEM.NPM, package, version)

        # For now, automatically allow all non-`install` commands
        if not self._is_install_command():
            return []

        # The presence of these options prevent the install command from running
        if any(opt in self._command for opt in {"-h", "--help", "--dry-run"}):
            return []

        try:
            # Compute the set of dependencies added by the install command
            dry_run_command = self._command + ["--dry-run", "--loglevel", "silly"]
            dry_run = subprocess.run(dry_run_command, check=True, text=True, capture_output=True)
            dependencies = map(line_to_dependency, filter(is_place_dep_line, dry_run.stderr.strip().split('\n')))
        except subprocess.CalledProcessError:
            # An erroring command does not install anything
            _log.info("The npm command encountered an error while collecting installation targets")
            return []

        try:
            # List targets already installed in the npm environment
            list_command = [self._executable, "list", "--all"]
            installed = subprocess.run(list_command, check=True, text=True, capture_output=True).stdout
        except subprocess.CalledProcessError:
            # If this operation fails, rather than blocking, assume nothing is installed
            # This has the effect of treating all dependencies like installation targets
            _log.warning(
                "Failed to list installed npm packages: treating all dependencies as installation targets"
            )
            installed = ""

        # The installation targets are the dependencies that are not already installed
        targets = filter(lambda dep: dep not in installed, dependencies)

        return list(map(str_to_install_target, targets))

    def _is_install_command(self) -> bool:
        """
        Determine whether the underlying `npm` command is for an `install` subcommand.

        Returns:
            A `bool` indicating whether the `npm` command underlying the given `NpmCommand`
            is likely for an `install` subcommand.

            This function gives no false negatives but may give false positives. False
            positives are safe in this case because they result in non-installish
            commands being analyzed as if they were installish commands. To eliminate
            false positives, we would need to write a full parser for npm.
        """
        # https://docs.npmjs.com/cli/v10/commands/npm-install
        install_aliases = {
            "install", "add", "i", "in", "ins", "inst", "insta", "instal", "isnt", "isnta", "isntal", "isntall"
        }

        return any(alias in self._command for alias in install_aliases)



"""
Provides utilities for configuring the environment (via `.rc` files) for using Supply-Chain Firewall.
"""

from pathlib import Path
import re
import os
import tempfile

from scfw.configure.constants import DD_AGENT_PORT_VAR, DD_API_KEY_VAR, DD_LOG_LEVEL_VAR

_CONFIG_FILES = [".bashrc", ".zshrc"]

_BLOCK_START = "# BEGIN SCFW MANAGED BLOCK"
_BLOCK_END = "# END SCFW MANAGED BLOCK"


def update_config_files(answers: dict):
    """
    Update the firewall's configuration in all supported .rc files.

    Args:
        answers: A `dict` configuration options to format and write to each file.
    """
    for file in [Path.home() / file for file in _CONFIG_FILES]:
        if file.exists():
            _update_config_file(file, answers)


def _update_config_file(config_file: Path, answers: dict):
    """
    Update the firewall's section in the given configuration file.

    Args:
        config_file: A `Path` to the configuration file to update.
        answers: The `dict` of configuration options to write.
    """
    def enclose(config: str) -> str:
        return f"{_BLOCK_START}{config}\n{_BLOCK_END}"

    with open(config_file) as f:
        contents = f.read()

    config = _format_answers(answers)

    pattern = f"{_BLOCK_START}(.*?){_BLOCK_END}"
    if not config:
        pattern = f"\n{pattern}\n"

    updated = re.sub(pattern, enclose(config) if config else '', contents, flags=re.DOTALL)
    if updated == contents and config not in contents:
        updated = f"{contents}\n{enclose(config)}\n"

    temp_fd, temp_file = tempfile.mkstemp(text=True)
    temp_handle = os.fdopen(temp_fd, 'w')
    temp_handle.write(updated)
    temp_handle.close()

    os.rename(temp_file, config_file)


def _format_answers(answers: dict) -> str:
    """
    Format configuration options into .rc file `str` content.

    Args:
        answers: A `dict` containing the user's selected configuration options.

    Returns:
        A `str` containing the desired configuration content for writing into a .rc file.
    """
    config = ''

    if answers.get("alias_pip"):
        config += '\nalias pip="scfw run pip"'
    if answers.get("alias_npm"):
        config += '\nalias npm="scfw run npm"'
    if (dd_agent_port := answers.get("dd_agent_port")):
        config += f'\nexport {DD_AGENT_PORT_VAR}="{dd_agent_port}"'
    if (dd_api_key := answers.get("dd_api_key")):
        config += f'\nexport {DD_API_KEY_VAR}="{dd_api_key}"'
    if (dd_log_level := answers.get("dd_log_level")):
        config += f'\nexport {DD_LOG_LEVEL_VAR}="{dd_log_level}"'

    return config



"""
Defines an installation target verifier that uses Datadog Security Research's
malicious software packages dataset.
"""

import requests

from scfw.ecosystem import ECOSYSTEM
from scfw.target import InstallTarget
from scfw.verifier import FindingSeverity, InstallTargetVerifier

_DD_DATASET_SAMPLES_URL = "https://raw.githubusercontent.com/DataDog/malicious-software-packages-dataset/main/samples"


class DatadogMaliciousPackagesVerifier(InstallTargetVerifier):
    """
    An `InstallTargetVerifier` for Datadog Security Research's malicious packages dataset.
    """
    def __init__(self):
        """
        Initialize a new `DatadogMaliciousPackagesVerifier`.

        Raises:
            requests.HTTPError: An error occurred while fetching a manifest file.
        """
        def download_manifest(ecosystem: str) -> dict[str, list[str]]:
            manifest_url = f"{_DD_DATASET_SAMPLES_URL}/{ecosystem}/manifest.json"
            request = requests.get(manifest_url, timeout=5)
            request.raise_for_status()
            return request.json()

        self._pypi_manifest = download_manifest("pypi")
        self._npm_manifest = download_manifest("npm")

    def name(self) -> str:
        """
        Return the `DatadogMaliciousPackagesVerifier` name string.

        Returns:
            The class' constant name string: `"DatadogMaliciousPackagesVerifier"`.
        """
        return "DatadogMaliciousPackagesVerifier"

    def verify(self, target: InstallTarget) -> list[tuple[FindingSeverity, str]]:
        """
        Determine whether the given installation target is malicious by consulting
        the dataset's manifests.

        Args:
            target: The installation target to verify.

        Returns:
            A list containing any findings for the given installation target, obtained
            by checking for its presence in the dataset's manifests.  Only a single
            `CRITICAL` finding to this effect is present in this case.
        """
        match target.ecosystem:
            case ECOSYSTEM.PIP:
                manifest = self._pypi_manifest
            case ECOSYSTEM.NPM:
                manifest = self._npm_manifest

        # We take the more conservative approach of ignoring version numbers when
        # deciding whether the given target is malicious
        if target.package in manifest:
            return [
                (
                    FindingSeverity.CRITICAL,
                    f"Datadog Security Research has determined that package {target.package} is malicious"
                )
            ]
        else:
            return []


def load_verifier() -> InstallTargetVerifier:
    """
    Export `DatadogMaliciousPackagesVerifier` for discovery by the firewall.

    Returns:
        A `DatadogMaliciousPackagesVerifier` for use in a run of the supply chain firewall.
    """
    return DatadogMaliciousPackagesVerifier()



"""
Exports the currently discoverable set of installation target verifiers
for use in the supply chain firewall's main routine.

Two installation target verifiers ship with the supply chain firewall
by default: one that uses Datadog Security Research's public malicious
packages dataset and one that uses OSV.dev. Users of the supply chain
firewall may additionally provide custom verifiers representing alternative
sources of truth for the firewall to use.

The firewall discovers verifiers at runtime via the following simple protocol.
The module implementing the custom verifier must contain a function with the
following name and signature:

```
def load_verifier() -> InstallTargetVerifier
```

This `load_verifier` function should return an instance of the custom verifier
for the firewall's use. The module may then be placed in the same directory
as this source file for runtime import. Make sure to reinstall the package
after doing so.
"""

import importlib
import logging
import os
import pkgutil

from scfw.verifier import InstallTargetVerifier

_log = logging.getLogger(__name__)


def get_install_target_verifiers() -> list[InstallTargetVerifier]:
    """
    Return the currently discoverable set of installation target verifiers.

    Returns:
        A list of the discovered installation target verifiers.
    """
    verifiers = []

    for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
        try:
            verifier = importlib.import_module(f".{module}", package=__name__).load_verifier()
            verifiers.append(verifier)
        except ModuleNotFoundError:
            _log.warning(f"Failed to load module {module} while collecting installation target verifiers")
        except AttributeError:
            _log.warning(f"Module {module} does not export an installation target verifier")

    return verifiers



"""
Provides the supply-chain firewall's main routine.
"""

import logging
import time

import scfw.cli as cli
from scfw.cli import Subcommand
import scfw.configure as configure
import scfw.firewall as firewall

_log = logging.getLogger(__name__)


def main() -> int:
    """
    The supply-chain firewall's main routine.

    Returns:
        An integer status code, 0 or 1.
    """
    args, help = cli.parse_command_line()

    if not args:
        print(help, end='')
        return 0

    _configure_logging(args.log_level)

    _log.info(f"Starting Supply-Chain Firewall on {time.asctime(time.localtime())}")
    _log.debug(f"Command line: {vars(args)}")

    match Subcommand(args.subcommand):
        case Subcommand.Configure:
            return configure.run_configure(args)
        case Subcommand.Run:
            return firewall.run_firewall(args)

    return 0


def _configure_logging(level: int):
    """
    Configure the root logger.

    Args:
        level: The log level selected by the user.
    """
    handler = logging.StreamHandler()
    handler.addFilter(logging.Filter(name="scfw"))
    handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))

    log = logging.getLogger()
    log.addHandler(handler)
    log.setLevel(level)



"""
Exports the currently discoverable set of client loggers implementing the
firewall's logging protocol.

One logger ships with the supply chain firewall by default: `DDLogger`,
which sends logs to Datadog. Firewall users may additionally provide
custom loggers according to their own logging needs.

The firewall discovers loggers at runtime via the following simple protocol.
The module implementing the custom logger must contain a function with the
following name and signature:

```
def load_logger() -> FirewallLogger
```

This `load_logger` function should return an instance of the custom logger
for the firewall's use. The module may then be placed in the same directory
as this source file for runtime import. Make sure to reinstall the package
after doing so.
"""

import importlib
import logging
import os
import pkgutil

from scfw.logger import FirewallLogger

_log = logging.getLogger(__name__)


def get_firewall_loggers() -> list[FirewallLogger]:
    """
    Return the currently discoverable set of client loggers.

    Returns:
        A `list` of the discovered `FirewallLogger`s.
    """
    loggers = []

    for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
        try:
            logger = importlib.import_module(f".{module}", package=__name__).load_logger()
            loggers.append(logger)
        except ModuleNotFoundError:
            _log.warning(f"Failed to load module {module} while collecting loggers")
        except AttributeError:
            _log.info(f"Module {module} does not export a logger")

    return loggers



"""
Provides a base class for installation target verifiers.
"""

from abc import (ABCMeta, abstractmethod)
from enum import Enum

from scfw.target import InstallTarget


class FindingSeverity(Enum):
    """
    A hierarchy of severity levels for installation target verifier findings.

    Installation target verifiers attach severity levels to their findings in
    order to direct the supply-chain firewall to take the correct action with
    respect to blocking or warning on an installation request.

    A `CRITICAL` finding causes the supply-chain firewall to block. A `WARNING`
    finding prompts the firewall to seek confirmation from the user before
    proceeding with the installation request.
    """
    CRITICAL = "CRITICAL"
    WARNING = "WARNING"


class InstallTargetVerifier(metaclass=ABCMeta):
    """
    Abstract base class for installation target verifiers.

    Each installation target verifier should implement a service for verifying
    installation targets in all supported ecosystems against a single reputable
    source of data on vulnerable and malicious open source packages.
    """
    @abstractmethod
    def name(self) -> str:
        """
        Return the verifier's name.

        Returns:
            A constant, short, descriptive name `str` identifying the verifier.
        """
        pass

    @abstractmethod
    def verify(self, target: InstallTarget) -> list[tuple[FindingSeverity, str]]:
        """
        Verify the given installation target.

        Args:
            target: The installation target to verify.

        Returns:
            A `list[tuple[FindingSeverity, str]]` of all findings for the given
            installation target reported by the backing data source, each tagged
            with a severity level for the firewall's use.

            Each `str` in this list should be a concise summary of a single finding
            and would ideally provide a link or handle to more information about that
            finding for the benefit of the user.
        """
        pass



"""
A representation of package ecosystems supported by the supply-chain firewall.
"""

from enum import Enum


class ECOSYSTEM(Enum):
    """
    Package ecosystems supported by the supply-chain firewall.
    """
    PIP = "pip"
    NPM = "npm"

    def __str__(self) -> str:
        """
        Format an `ECOSYSTEM` for printing.

        Returns:
            A `str` representing the given `ECOSYSTEM` suitable for printing.
        """
        return self.value



"""
Provides utilities for configuring the local Datadog Agent to receive logs from Supply-Chain Firewall.
"""

import json
from pathlib import Path
import shutil
import subprocess

from scfw.configure.constants import DD_SERVICE, DD_SOURCE


def configure_agent_logging(port: str):
    """
    Configure a local Datadog Agent for accepting logs from the firewall.

    Args:
        port: The local port number where the firewall logs will be sent to the Agent.

    Raises:
        ValueError: An invalid port number was provided.
        RuntimeError: An error occurred while querying the Agent's status.
    """
    if not (0 < int(port) < 65536):
        raise ValueError("Invalid port number provided for Datadog Agent logging")

    config_file = (
        "logs:\n"
        "  - type: tcp\n"
        f"    port: {port}\n"
        f'    service: "{DD_SERVICE}"\n'
        f'    source: "{DD_SOURCE}"\n'
    )

    scfw_config_dir = _dd_agent_scfw_config_dir()
    scfw_config_file = scfw_config_dir / "conf.yaml"

    if not scfw_config_dir.is_dir():
        scfw_config_dir.mkdir()
    with open(scfw_config_file, 'w') as f:
        f.write(config_file)


def remove_agent_logging():
    """
    Remove Datadog Agent configuration for Supply-Chain Firewall, if it exists.

    Raises:
        RuntimeError: An error occurred while attempting to remove the configuration directory.
    """
    scfw_config_dir = _dd_agent_scfw_config_dir()

    if not scfw_config_dir.is_dir():
        return

    try:
        shutil.rmtree(scfw_config_dir)
    except Exception:
        raise RuntimeError(
            "Failed to delete Datadog Agent configuration directory for Supply-Chain Firewall"
        )


def _dd_agent_scfw_config_dir() -> Path:
    """
    Get the filesystem path to the firewall's configuration directory for
    Datadog Agent log forwarding.

    Returns:
        A `Path` indicating the absolute filesystem path to this directory.

    Raises:
        RuntimeError:
            Unable to query Datadog Agent status to read the location of its
            global configuration directory.
    """
    try:
        agent_status = subprocess.run(
            ["datadog-agent", "status", "--json"], check=True, text=True, capture_output=True
        )
        agent_config_dir = json.loads(agent_status.stdout).get("config", {}).get("confd_path", "")

    except subprocess.CalledProcessError:
        raise RuntimeError(
            "Unable to query Datadog Agent status: please ensure the Agent is running. "
            "Linux users may need sudo to run this command."
        )

    return Path(agent_config_dir) / "scfw.d"



"""
Provides various configuration-related constants.
"""

DD_SOURCE = "scfw"
"""
Source value for Datadog logging.
"""

DD_SERVICE = "scfw"
"""
Service value for Datadog logging.
"""

DD_ENV = "dev"
"""
Default environment value for Datadog logging.
"""

DD_API_KEY_VAR = "DD_API_KEY"
"""
The environment variable under which the firewall looks for a Datadog API key.
"""

DD_LOG_LEVEL_VAR = "SCFW_DD_LOG_LEVEL"
"""
The environment variable under which the firewall looks for a Datadog log level setting.
"""

DD_AGENT_PORT_VAR = "SCFW_DD_AGENT_LOG_PORT"
"""
The environment variable under which the firewall looks for a port number on which to
forward firewall logs to the local Datadog Agent.
"""



"""
Configures a logger for sending firewall logs to Datadog's API over HTTP.
"""

import logging
import os
import socket

from datadog_api_client import ApiClient, Configuration
from datadog_api_client.v2.api.logs_api import LogsApi
from datadog_api_client.v2.model.content_encoding import ContentEncoding
from datadog_api_client.v2.model.http_log import HTTPLog
from datadog_api_client.v2.model.http_log_item import HTTPLogItem

import scfw
from scfw.configure import DD_API_KEY_VAR, DD_ENV, DD_SERVICE, DD_SOURCE
from scfw.logger import FirewallLogger
from scfw.loggers.dd_logger import DDLogFormatter, DDLogger

_DD_LOG_NAME = "dd_api_log"


class _DDLogHandler(logging.Handler):
    """
    A log handler for adding tags and forwarding firewall logs of blocked and
    permitted package installation requests to the Datadog API.

    In addition to USM tags, install targets are tagged with the `target` tag and included.
    """
    def emit(self, record):
        """
        Format and send a log to Datadog.

        Args:
            record: The log record to be forwarded.
        """
        usm_tags = {
            f"env:{os.getenv('DD_ENV', DD_ENV)}",
            f"version:{scfw.__version__}"
        }

        targets = record.__dict__.get("targets", {})
        target_tags = set(map(lambda e: f"target:{e}", targets))

        body = HTTPLog(
            [
                HTTPLogItem(
                    ddsource=DD_SOURCE,
                    ddtags=",".join(usm_tags | target_tags),
                    hostname=socket.gethostname(),
                    message=self.format(record),
                    service=DD_SERVICE,
                ),
            ]
        )

        configuration = Configuration()
        with ApiClient(configuration) as api_client:
            api_instance = LogsApi(api_client)
            api_instance.submit_log(content_encoding=ContentEncoding.DEFLATE, body=body)


# Configure a single logging handle for all `DDAPILogger` instances to share
_handler = _DDLogHandler() if os.getenv(DD_API_KEY_VAR) else logging.NullHandler()
_handler.setFormatter(DDLogFormatter())

_ddlog = logging.getLogger(_DD_LOG_NAME)
_ddlog.setLevel(logging.INFO)
_ddlog.addHandler(_handler)


class DDAPILogger(DDLogger):
    """
    An implementation of `FirewallLogger` for sending logs to the Datadog API.
    """
    def __init__(self):
        """
        Initialize a new `DDAPILogger`.
        """
        super().__init__(_ddlog)


def load_logger() -> FirewallLogger:
    """
    Export `DDAPILogger` for discovery by the firewall.

    Returns:
        A `DDAPILogger` for use in a run of the firewall.
    """
    return DDAPILogger()



"""
Defines a subclass of `PackageManagerCommand` for `pip` commands.
"""

import json
import logging
import os
import subprocess
import sys
from typing import Optional

from packaging.version import InvalidVersion, Version, parse as version_parse

from scfw.command import PackageManagerCommand, UnsupportedVersionError
from scfw.ecosystem import ECOSYSTEM
from scfw.target import InstallTarget

_log = logging.getLogger(__name__)

MIN_PIP_VERSION = version_parse("22.2")

_UNSUPPORTED_PIP_VERSION = f"pip before v{MIN_PIP_VERSION} is not supported"


class PipCommand(PackageManagerCommand):
    """
    A representation of `pip` commands via the `PackageManagerCommand` interface.
    """
    def __init__(self, command: list[str], executable: Optional[str] = None):
        """
        Initialize a new `PipCommand`.

        Args:
            command: A `pip` command line.
            executable:
                Optional path to the executable to run the command.  Determined by the
                environment if not given.

        Raises:
            ValueError: An invalid `pip` command line was given.
            UnsupportedVersionError:
                An unsupported version of `pip` was used to initialize a `PipCommand`.
        """
        def get_executable() -> str:
            if (venv := os.environ.get("VIRTUAL_ENV")):
                return os.path.join(venv, "bin/python")
            else:
                return sys.executable

        def get_pip_version(executable: str) -> Version:
            try:
                # All supported versions adhere to this format
                pip_version_command = [executable, "-m", "pip", "--version"]
                pip_version = subprocess.run(pip_version_command, check=True, text=True, capture_output=True)
                version_str = pip_version.stdout.split()[1]
                return version_parse(version_str)
            except IndexError:
                raise UnsupportedVersionError(_UNSUPPORTED_PIP_VERSION)
            except InvalidVersion:
                raise UnsupportedVersionError(_UNSUPPORTED_PIP_VERSION)

        if not command or command[0] != "pip":
            raise ValueError("Malformed pip command")
        self._command = command

        self._executable = executable if executable else get_executable()
        if get_pip_version(self._executable) < MIN_PIP_VERSION:
            raise UnsupportedVersionError(_UNSUPPORTED_PIP_VERSION)

    def run(self):
        """
        Run a `pip` command.
        """
        subprocess.run([self._executable, "-m"] + self._command)

    def would_install(self) -> list[InstallTarget]:
        """
        Determine the list of Python packages a `pip` command would install if it were run.

        Returns:
            A `list[InstallTarget]` representing the packages the `pip` command would
            install if it were run.

        Raises:
            ValueError: The `pip` install report did not have the required format.
        """
        def report_to_install_targets(install_report: dict) -> InstallTarget:
            if not (metadata := install_report.get("metadata")):
                raise ValueError("Missing metadata for pip install target")
            if not (package := metadata.get("name")):
                raise ValueError("Missing name for pip install target")
            if not (version := metadata.get("version")):
                raise ValueError("Missing version for pip install target")
            return InstallTarget(ECOSYSTEM.PIP, package, version)

        # pip only installs or upgrades packages via the `pip install` subcommand
        # If `install` is not present, the command is automatically safe to run
        # If `install` is present with any of the below options, a usage or error
        # message is printed or a dry-run install occurs: nothing will be installed
        if "install" not in self._command or any(opt in self._command for opt in {"-h", "--help", "--dry-run"}):
            return []

        # Otherwise, this is probably a "live" `pip install` command
        # To be certain, we would need to write a full parser for pip
        dry_run_command = [self._executable, "-m"] + self._command + ["--dry-run", "--quiet", "--report", "-"]
        try:
            dry_run = subprocess.run(dry_run_command, check=True, text=True, capture_output=True)
            install_report = json.loads(dry_run.stdout).get("install", [])
            return list(map(report_to_install_targets, install_report))
        except subprocess.CalledProcessError:
            # An error must have resulted from the given pip command
            # As nothing will be installed in this case, allow the command
            _log.info("The pip command encountered an error while collecting installation targets")
            return []



from scfw.cli import _parse_command_line, _DEFAULT_LOG_LEVEL


def test_cli_basic_usage_pip():
    """
    Basic pip command usage.
    """
    argv = ["scfw", "run", "pip", "install", "requests"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[2:]
    assert not args.dry_run
    assert not args.executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_basic_usage_npm():
    """
    Basic npm command usage.
    """
    argv = ["scfw", "run", "npm", "install", "react"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[2:]
    assert not args.dry_run
    assert not args.executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_no_options_no_command():
    """
    Invocation with no options or arguments.
    """
    argv = ["scfw"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_all_options_no_command():
    """
    Invocation with all top-level options and no subcommand.
    """
    argv = ["scfw", "--log-level", "DEBUG"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_incorrect_subcommand():
    """
    Invocation with a nonexistent subcommand.
    """
    argv = ["scfw", "nonexistent"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_all_options_no_command():
    """
    Invocation with all options and no arguments.
    """
    executable = "/usr/bin/python"
    argv = ["scfw", "run", "--executable", executable, "--dry-run"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_all_options_pip():
    """
    Invocation with all options and a pip command argument.
    """
    executable = "/usr/bin/python"
    argv = ["scfw", "run", "--executable", executable, "--dry-run", "pip", "install", "requests"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[5:]
    assert args.dry_run
    assert args.executable == executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_all_options_npm():
    """
    Invocation with all options and an npm command argument.
    """
    executable = "/opt/homebrew/bin/npm"
    argv = ["scfw", "run", "--executable", executable, "--dry-run", "npm", "install", "react"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[5:]
    assert args.dry_run
    assert args.executable == executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_package_manager_dry_run_pip():
    """
    Test that a pip `--dry-run` flag is parsed correctly as such.
    """
    argv = ["scfw", "run", "pip", "--dry-run", "install", "requests"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[2:]
    assert not args.dry_run
    assert not args.executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_package_manager_dry_run_npm():
    """
    Test that an npm `--dry-run` flag is parsed correctly as such.
    """
    argv = ["scfw", "run", "npm", "--dry-run", "install", "react"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "run"
    assert args.command == argv[2:]
    assert not args.dry_run
    assert not args.executable
    assert args.log_level == _DEFAULT_LOG_LEVEL


def test_cli_pip_over_npm():
    """
    Test that a pip command is parsed correctly in the presence of an "npm" literal.
    """
    argv = ["scfw", "run", "pip", "install", "npm"]
    args, _ = _parse_command_line(argv)
    assert args.command == argv[2:]


def test_cli_npm_over_pip():
    """
    Test that an npm command is parsed correctly in the presence of a "pip" literal.
    """
    argv = ["scfw", "run", "npm", "install", "pip"]
    args, _ = _parse_command_line(argv)
    assert args.command == argv[2:]


def test_cli_basic_usage_configure():
    """
    Basic `configure` subcommand usage.
    """
    argv = ["scfw", "configure"]
    args, _ = _parse_command_line(argv)
    assert args.subcommand == "configure"
    assert "command" not in args
    assert "dry_run" not in args
    assert "executable" not in args
    assert args.log_level == _DEFAULT_LOG_LEVEL



"""
A representation of package manager installation targets in a supported ecosystem.
"""

from dataclasses import dataclass

from scfw.ecosystem import ECOSYSTEM


@dataclass(eq=True, frozen=True)
class InstallTarget:
    """
    An installation target in a particular ecosystem.

    Attributes:
        ecosystem: The ecosystem of the installation target.
        package: The installation target's package name in its ecosystem.
        version: The installation target's version number.
    """
    ecosystem: ECOSYSTEM
    package: str
    version: str

    def __str__(self) -> str:
        """
        Format the `InstallTarget` package and version number as a string according
        to the conventions of its ecosystem.

        Returns:
            A `str` with ecosystem-specific formatting describing the `InstallTarget`
            package and version number.

            For `pip` packages, the formatting used is `"{package}-{version}"` and for `npm`
            packages it is `"{package}@{version}"`.
        """
        match self.ecosystem:
            case ECOSYSTEM.PIP:
                return f"{self.package}-{self.version}"
            case ECOSYSTEM.NPM:
                return f"{self.package}@{self.version}"



"""
Defines an installation target verifier that uses OSV.dev's database of vulnerable
and malicious open source software packages.
"""

import logging

import requests

from scfw.ecosystem import ECOSYSTEM
from scfw.target import InstallTarget
from scfw.verifier import FindingSeverity, InstallTargetVerifier

_log = logging.getLogger(__name__)

_OSV_ECOSYSTEMS = {ECOSYSTEM.PIP: "PyPI", ECOSYSTEM.NPM: "npm"}

_OSV_DEV_QUERY_URL = "https://api.osv.dev/v1/query"
_OSV_DEV_VULN_URL_PREFIX = "https://osv.dev/vulnerability"
_OSV_DEV_LIST_URL_PREFIX = "https://osv.dev/list"


class OsvVerifier(InstallTargetVerifier):
    """
    An `InstallTargetVerifier` for the OSV.dev open source vulnerability and
    malicious package database.
    """
    def name(self) -> str:
        """
        Return the `OsvVerifier` name string.

        Returns:
            The class' constant name string: `"OsvVerifier"`.
        """
        return "OsvVerifier"

    def verify(self, target: InstallTarget) -> list[tuple[FindingSeverity, str]]:
        """
        Query an given installation target against the OSV.dev database.

        Args:
            target: The installation target to query.

        Returns:
            A list containing any findings for the given installation target, obtained
            by querying for it against OSV.dev.

            OSV.dev disclosures with `MAL` IDs are treated as `CRITICAL` findings and all
            others are treated as `WARNING`.  *It is very important to note that most but
            **not all** OSV.dev malicious package disclosures have `MAL` IDs.*

        Raises:
            requests.HTTPError:
                An error occurred while querying an installation target against the OSV.dev API.
        """
        def mal_finding(id: str) -> str:
            return (
                f"An OSV.dev malicious package disclosure exists for package {target}:\n"
                f"  * {_OSV_DEV_VULN_URL_PREFIX}/{id}"
            )

        def non_mal_finding(id: str) -> str:
            return (
                f"An OSV.dev disclosure exists for package {target}:\n"
                f"  * {_OSV_DEV_VULN_URL_PREFIX}/{id}"
            )

        def error_message(e: str) -> str:
            url = f"{_OSV_DEV_LIST_URL_PREFIX}?q={target.package}&ecosystem={_OSV_ECOSYSTEMS[target.ecosystem]}"
            return (
                f"Failed to verify target against OSV.dev: {e if e else 'An unspecified error occurred'}.\n"
                f"Before proceeding, please check for OSV.dev advisories related to this target.\n"
                f"DO NOT PROCEED if it has an advisory with a MAL ID: it is very likely malicious.\n"
                f"  * {url}"
            )

        query = {
            "version": target.version,
            "package": {
                "name": target.package,
                "ecosystem": _OSV_ECOSYSTEMS[target.ecosystem]
            }
        }

        try:
            # The OSV.dev API is sometimes quite slow, hence the generous timeout
            request = requests.post(_OSV_DEV_QUERY_URL, json=query, timeout=10)
            request.raise_for_status()

            if not (vulns := request.json().get("vulns")):
                return []

            osv_ids = set(filter(lambda id: id is not None, map(lambda vuln: vuln.get("id"), vulns)))
            mal_ids = set(filter(lambda id: id.startswith("MAL"), osv_ids))
            non_mal_ids = osv_ids - mal_ids

            return (
                [(FindingSeverity.CRITICAL, mal_finding(id)) for id in mal_ids]
                + [(FindingSeverity.WARNING, non_mal_finding(id)) for id in non_mal_ids]
            )

        except requests.exceptions.RequestException as e:
            _log.warning(f"Failed to query OSV.dev API: returning WARNING finding for target {target}")
            return [(FindingSeverity.WARNING, error_message(str(e)))]


def load_verifier() -> InstallTargetVerifier:
    """
    Export `OsvVerifier` for discovery by the firewall.

    Returns:
        An `OsvVerifier` for use in a run of the supply chain firewall.
    """
    return OsvVerifier()



"""
Provides the core orchestration logic for verifying a set of installation targets
and structuring and printing the results.
"""

import concurrent.futures as cf
import itertools
import logging
from typing import TypeAlias

from scfw.target import InstallTarget
from scfw.verifier import FindingSeverity, InstallTargetVerifier

_log = logging.getLogger(__name__)


VerificationReport: TypeAlias = dict[InstallTarget, list[str]]
"""A structured report containing findings resulting from installation target verification."""


def verify_install_targets(
    verifiers: list[InstallTargetVerifier],
    targets: list[InstallTarget]
) -> dict[FindingSeverity, VerificationReport]:
    """
    Verify a set of installation targets against a set of verifiers.

    Args:
        verifiers: The set of verifiers to use against the installation targets.
        targers: A list of installation targets to be verified.

    Returns:
        A `dict[FindingSeverity, VerificationReport]` representing the severity-ranked
        results of verification across all installation targets and verifiers.

        The returned `dict` is such that a given `FindingSeverity` key is present iff
        its `VerificationReport` value has a finding for some installation target.
        Moreover, this finding was determined to be at that severity level by the
        verifier that returned it.
    """
    reports: dict[FindingSeverity, VerificationReport] = {}

    with cf.ThreadPoolExecutor() as executor:
        task_results = {
            executor.submit(lambda v, t: v.verify(t), verifier, target): (verifier.name(), target)
            for verifier, target in itertools.product(verifiers, targets)
        }
        for future in cf.as_completed(task_results):
            verifier, target = task_results[future]
            if (findings := future.result()):
                _log.info(f"Verifier {verifier} had findings for target {target}")
                for severity, finding in findings:
                    if severity not in reports:
                        reports[severity] = {}
                    if target not in reports[severity]:
                        reports[severity][target] = [finding]
                    else:
                        reports[severity][target].append(finding)
            else:
                _log.info(f"Verifier {verifier} had no findings for target {target}")

    _log.info("Verification of installation targets complete")
    return reports


def show_verification_report(report: VerificationReport) -> str:
    """
    Return a human-readable version of a verification report.

    Returns:
        A `str` containing the formatted verification report.
    """
    def show_line(linenum: int, line: str) -> str:
        return (f"  - {line}" if linenum == 0 else f"    {line}")

    def show_finding(finding: str) -> str:
        return '\n'.join(
            show_line(linenum, line) for linenum, line in enumerate(finding.split('\n'))
        )

    def show_findings(target: InstallTarget, findings: list[str]) -> str:
        return (
            f"Installation target {target}:\n" + '\n'.join(map(show_finding, findings))
        )

    return '\n'.join(
        show_findings(target, findings) for target, findings in report.items()
    )



"""
A drop-in replacement for `argparse.ArgumentParser`.
"""

import argparse


class ArgumentParser(argparse.ArgumentParser):
    """
    A drop-in replacement for `argparse.ArgumentParser` with a patched
    implementation of the latter's `exit_on_error` behavior.

    See https://github.com/python/cpython/issues/103498 for more info.
    """
    def error(self, message):
        """
        Handle a parsing error.

        Args:
            message: The error message.
        """
        raise argparse.ArgumentError(None, message)



"""
Provides a base class for representing the commands of package managers like `pip` and `npm`.
"""

from abc import (ABCMeta, abstractmethod)
from typing import Optional

from scfw.target import InstallTarget


class PackageManagerCommand(metaclass=ABCMeta):
    """
    Abstract base class for commands in an ecosystem's package manager.
    """
    @abstractmethod
    def __init__(self, command: list[str], executable: Optional[str] = None):
        """
        Initialize a new package manager command.

        Args:
            command: The package manager command line as provided to the supply-chain firewall.
            executable:
                Optional path to the executable to run the command.  Determined by the environment
                where the firewall is running if not given.

        Raises:
            UnsupportedVersionError:
                Subclasses should raise this error when an attempt is made to initialize an
                instance with an unsupported version of the underlying package manager.
        """
        pass

    @abstractmethod
    def run(self):
        """
        Run a package manager command.
        """
        pass

    @abstractmethod
    def would_install(self) -> list[InstallTarget]:
        """
        Without running the command, determine the packages that would be installed by a
        package manager command if it were run.

        Returns:
            A list of `InstallTarget` representing the installation targets the command would
            install or upgrade if it were run.
        """
        pass


class UnsupportedVersionError(Exception):
    """
    An error that occurs when an attempt is made to initialize a `PackageManagerCommand`
    with an unsupported version of the underlying package manager.

    When this occurs, subclasses of `PackageManagerCommand` should raise
    `UnsupportedVersionError` in their `__init__()` methods.  It should not be possible
    to initialize a `PackageManagerCommand` with an unsupported version.  The error
    should contain a message to the user with instructions for how to upgrade their
    package manager to a supported version.

    When `UnsupportedVersionError` is raised, the firewall logs an error message stating
    that an unsupported version error occurred, logs the error message from the
    `PackageManagerCommand`, and exits gracefully.
    """
    pass



"""
Configures a logger for sending firewall logs to a local Datadog Agent.
"""

import logging
import os
import socket

from scfw.configure import DD_AGENT_PORT_VAR
from scfw.logger import FirewallLogger
from scfw.loggers.dd_logger import DDLogFormatter, DDLogger

_log = logging.getLogger(__name__)

_DD_LOG_NAME = "dd_agent_log"


class _DDLogHandler(logging.Handler):
    def emit(self, record):
        """
        Format and send a log to the Datadog Agent.

        Args:
            record: The log record to be forwarded.
        """
        try:
            message = self.format(record).encode()

            agent_port = int(os.getenv(DD_AGENT_PORT_VAR))

            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", agent_port))
            if s.send(message) != len(message):
                raise ValueError("Log send failed")
            s.close()

        except Exception as e:
            _log.warning(f"Failed to forward log to Datadog Agent: {e}")


# Configure a single logging handle for all `DDAgentLogger` instances to share
_handler = _DDLogHandler() if os.getenv(DD_AGENT_PORT_VAR) else logging.NullHandler()
_handler.setFormatter(DDLogFormatter())

_ddlog = logging.getLogger(_DD_LOG_NAME)
_ddlog.setLevel(logging.INFO)
_ddlog.addHandler(_handler)


class DDAgentLogger(DDLogger):
    """
    An implementation of `FirewallLogger` for sending logs to a local Datadog Agent.
    """
    def __init__(self):
        """
        Initialize a new `DDAgentLogger`.
        """
        super().__init__(_ddlog)


def load_logger() -> FirewallLogger:
    """
    Export `DDAgentLogger` for discovery by the firewall.

    Returns:
        A `DDAgentLogger` for use in a run of the firewall.
    """
    return DDAgentLogger()



"""
Provides an interface for obtaining `PackageManagerCommand` instances from given
command lines for supported package managers.
"""

from typing import Optional

from scfw.command import PackageManagerCommand
from scfw.commands.npm_command import NpmCommand
from scfw.commands.pip_command import PipCommand
from scfw.ecosystem import ECOSYSTEM


def get_package_manager_command(
    command: list[str],
    executable: Optional[str] = None
) -> tuple[ECOSYSTEM, PackageManagerCommand]:
    """
    Return a `PackageManagerCommand` for the given ecosystem and arguments.

    Args:
        command: The command line of the desired command as provided to the supply-chain firewall.
        executable: An optional executable to use when running the package manager command.

    Returns:
        A `tuple` of the `ECOSYSTEM` corresponding to the received command line and a
        `PackageManagerCommand` initialized from that command line.

    Raises:
        ValueError: An empty or unsupported package manager command line was provided.
    """
    if not command:
        raise ValueError("Missing package manager command")

    try:
        match (ecosystem := ECOSYSTEM(command[0])):
            case ECOSYSTEM.PIP:
                return ecosystem, PipCommand(command, executable)
            case ECOSYSTEM.NPM:
                return ecosystem, NpmCommand(command, executable)

    except ValueError:
        raise ValueError(f"Unsupported package manager '{command[0]}'")
